# SPDX-FileCopyrightText: 2023-present Alan Bridgeman <a.bridgeman@hotmail.com>
#
# SPDX-License-Identifier: MIT
from .AzureResources import AzureResources
from .AppRegistration import AppRegistration
from .ResourceGroup import ResourceGroup
from .WebApp import WebApp
from .ContainerRegistry import ContainerRegistry
from .Database import Database
from .RedisCache import RedisCache
from .StorageAccount import StorageAccount
from .StorageContainer import StorageContainer
from .StorageTable import StorageTable
from .KeyVault import KeyVault
from .VarName import VarName



from typing import Any

from .interfaces.IAppRegistration import IAppRegistration

class AppRegistration(IAppRegistration):
    def __init__(self, definition_json: dict[str, Any]):
        self.__definition_json = definition_json

        self.__name = definition_json[self.NAME_KEY]
        
        self._app_id = None
        self._app_secret = None
        self._service_principal_id = None
    
    @property
    def name(self) -> str:
        """The name of the app registration"""
        
        return self.__name
    
    @property
    def app_id(self) -> str:
        """The app id of the app registration"""
        
        return self._app_id # type: ignore
    
    @app_id.setter
    def app_id(self, value: str):
        self._app_id = value
    
    @property
    def app_secret(self) -> str:
        """The app secret of associated with the app registration"""
        
        return self._app_secret # type: ignore
    
    @app_secret.setter
    def app_secret(self, value: str):
        self._app_secret = value
    
    @property
    def service_principal_id(self) -> str:
        """The id of the service principal associated with the app registration"""
        
        return self._service_principal_id # type: ignore
    
    @service_principal_id.setter
    def service_principal_id(self, value: str):
        self._service_principal_id = value
    
    def to_json(self) -> dict[str, Any]:
        return {
            self.NAME_KEY: self.name,
            "app_id": self.app_id,
            "app_secret": self.app_secret,
            "service_principal_id": self.service_principal_id
        }


from abc import ABCMeta, abstractmethod

class IVarNameStorageTable:
    __metaclass__ = ABCMeta

    @property
    @abstractmethod
    def category(self) -> str | None:
        """The category of the storage table"""
        
        pass
    
    @property
    @abstractmethod
    def name(self) -> str:
        """The name of the storage table"""
        
        pass
    
    @property
    @abstractmethod
    def var_name(self) -> str:
        """The variable name for the storage table"""
        
        pass


from typing import Any

from .interfaces.IResourceGroup import IResourceGroup

class ResourceGroup(IResourceGroup):
    """Represents a resource group in Azure
    
    Note, that properties are used with no setters. This is because the resource group definitions should not be changed.

    Attributes:
        name (str): The name of the resource group
    """

    def __init__(self, definition_json: dict[str, Any]):
        self.__definition_json = definition_json

        self.__name = definition_json[self.NAME_KEY]
    
    @property
    def name(self) -> str:
        """The name of the resource group"""
        
        return self.__name
    
    def to_json(self) -> dict[str, Any]:
        return {
            self.NAME_KEY: self.name
        }


from typing import Any

from .interfaces.IAzureResources import IAzureResources
from .interfaces.IWebApp import IWebApp

from .ResourceGroup import ResourceGroup

class WebApp(IWebApp):
    """Represents a webapp in Azure"""

    def __init__(self, definition_json: dict[str, Any], azure_resources: IAzureResources, category_name: str | None = None):
        self.__definition_json = definition_json
        
        self.__name = str(definition_json[self.NAME_KEY])
        self.__resource_group = self.__get_resource_group(definition_json, azure_resources)
        self.__category_name = category_name
        self.__app_service_plan = str(definition_json[self.APP_SERVICE_PLAN_KEY])
        self.__insights = bool(definition_json[self.INSIGHTS_KEY])
    
    def __get_resource_group(self, definition_json: dict[str, Any], azure_resources: IAzureResources) -> ResourceGroup:
        """Gets the resource group from the definition json
        
        Args:
            definition_json (dict): The definition json
            azure_resources (IAzureResources): The azure resources (used to get the resource group)
        
        Raises:
            Exception: If the resource group is not found
        
        Returns:
            ResourceGroup: The resource group
        """

        resource_group: ResourceGroup | None = azure_resources.get_resource_group(str(definition_json[self.RESOURCE_GROUP_KEY])) # type: ignore

        if resource_group is None:
            raise Exception(f"Resource group {str(definition_json[self.RESOURCE_GROUP_KEY])} not found")

        return resource_group

    @property
    def name(self) -> str:
        """The name of the webapp"""

        return self.__name
    
    @property
    def resource_group(self) -> ResourceGroup:
        """The resource group of the webapp"""

        return self.__resource_group
    
    @property
    def category_name(self) -> str | None:
        """The category name of the webapp"""

        return self.__category_name

    @property
    def app_service_plan(self) -> str:
        """The app service plan of the webapp"""

        return self.__app_service_plan
    
    @property
    def insights(self) -> bool:
        """If insights is enabled for the webapp"""

        return self.__insights
    
    def to_json(self) -> dict[str, Any]:
        return {
            self.NAME_KEY: self.name,
            self.RESOURCE_GROUP_KEY: self.resource_group.name,
            self.APP_SERVICE_PLAN_KEY: self.app_service_plan,
            self.INSIGHTS_KEY: self.insights
        }


from abc import ABCMeta, abstractmethod

class IResourceGroup:
    __metaclass__ = ABCMeta

    NAME_KEY = "name"
    """The key for the name of the resource group in the definition JSON"""

    @property
    @abstractmethod
    def name(self) -> str:
        """The name of the resource group"""
        pass


from abc import ABCMeta, abstractmethod

from .IResourceGroup import IResourceGroup
from .IWebApp import IWebApp
from .IContainerRegistry import IContainerRegistry
from .IAppRegistration import IAppRegistration
from .IDatabase import IDatabase
from .IRedisCache import IRedisCache
from .IStorageAccount import IStorageAccount
from .IKeyVault import IKeyVault
from .ICategorizableResource import ICategorizableResource

class IAzureResources:
    __metaclass__ = ABCMeta

    RESOURCE_GROUPS_KEY = 'resourceGroups'
    """Key for the resource groups section of an Azure Resources definition"""
    WEB_APPS_KEY = 'webApps'
    """Key for the web apps section of an Azure Resources definition"""
    CONTAINER_REGISTRIES_KEY = 'containerRegistries'
    """Key for the container registries section of an Azure Resources definition"""
    APP_REGISTRATIONS_KEY = 'appRegistrations'
    """Key for the app registrations section of an Azure Resources definition"""
    DATABASES_KEY = 'databases'
    """Key for the databases section of an Azure Resources definition"""
    CACHES_KEY = 'caches'
    """Key for the caches section of an Azure Resources definition"""
    STORAGE_KEY = 'storage'
    """Key for the storage section of an Azure Resources definition"""
    VAULTS_KEY = 'vaults'
    """Key for the vaults section of an Azure Resources definition"""
    CATEGORIZED_RESOURCES_KEY = 'categorizedResources'
    """Key for the categorized resources section of an Azure Resources definition"""

    @property
    @abstractmethod
    def resource_groups(self) -> list[IResourceGroup]:
        pass
    @abstractmethod
    def has_resource_group(self, name: str) -> bool:
        pass
    @abstractmethod
    def get_resource_group(self, name: str) -> IResourceGroup | None:
        pass
    
    @property
    @abstractmethod
    def web_apps(self) -> list[IWebApp]:
        pass
    @abstractmethod
    def has_web_app(self, name: str) -> bool:
        pass
    @abstractmethod
    def get_web_app(self, name: str) -> IWebApp | None:
        pass
    
    @property
    @abstractmethod
    def container_registries(self) -> list[IContainerRegistry]:
        pass
    @abstractmethod
    def has_container_registry(self, name: str) -> bool:
        pass
    @abstractmethod
    def get_container_registry(self, name: str) -> IContainerRegistry | None:
        pass
    
    @property
    @abstractmethod
    def app_registrations(self) -> list[IAppRegistration]:
        pass
    @abstractmethod    
    def has_app_registration(self, name: str) -> bool:
       pass
    @abstractmethod
    def get_app_registration(self, name: str) -> IAppRegistration | None:
        pass
    
    @property
    @abstractmethod
    def databases(self) -> list[IDatabase]:
        pass
    @abstractmethod
    def has_database(self, name: str) -> bool:
        pass
    @abstractmethod
    def get_database(self, name: str) -> IDatabase | None:
        pass

    @property
    @abstractmethod
    def caches(self) -> list[IRedisCache]:
        pass
    @abstractmethod
    def has_cache(self, name: str) -> bool:
        pass
    @abstractmethod
    def get_cache(self, name: str) -> IRedisCache | None:
        pass

    @property
    @abstractmethod
    def uncategorized_storage(self) -> list[IStorageAccount]:
        pass
    @abstractmethod
    def has_uncategorized_storage(self, name: str) -> bool:
        pass
    @abstractmethod
    def get_uncategorized_storage(self, name: str) -> IStorageAccount | None:
        pass

    @property
    @abstractmethod
    def uncategorized_vaults(self) -> list[IKeyVault]:
        pass
    @abstractmethod
    def has_uncategorized_vault(self, name: str) -> bool:
        pass
    @abstractmethod
    def get_uncategorized_vault(self, name: str) -> IKeyVault | None:
        pass

    
    @property
    @abstractmethod
    def categorized_resources(self) -> dict[str, dict[str, ICategorizableResource]]:
        pass
    @abstractmethod
    def get_resource_categories(self) -> list[str]:
        pass
    @abstractmethod
    def has_resource_category(self, category: str) -> bool:
        pass
    @abstractmethod    
    def get_category_resources(self, category: str) -> list[ICategorizableResource] | None:
        pass
    
    @classmethod
    @abstractmethod
    def from_file(cls, file_path: str):
        pass


from typing import Any

from .interfaces.IStorageContainer import IStorageContainer
from .interfaces.IVarNameStorageContainer import IVarNameStorageContainer
from .interfaces.IStorageAccount import IStorageAccount

from .VarName import VarName

class StorageContainer(IStorageContainer, IVarNameStorageContainer):
    def __init__(self, definition_json: dict[str, Any], account: IStorageAccount):
        self.__definition_json = definition_json

        self.__category = account.category

        self.__name = str(definition_json[self.NAME_KEY])
        self.__account = account
        self.__var_name = VarName(definition_json[self.VAR_NAME_KEY])
    
    @property
    def category(self) -> str | None:
        """The BA infrastructure category of the storage container"""

        return self.__category

    @property
    def name(self) -> str:
        """The name of the storage container"""
        
        return self.__name
    
    @property
    def account(self) -> IStorageAccount:
        """The storage account the storage container is in"""
        
        return self.__account
    
    @property
    def var_name_object(self) -> VarName:
        """The variable name (object) for the storage container"""
        return self.__var_name
    
    @property
    def var_name(self) -> str:
        """The variable name for the storage container"""
        return self.__var_name.get(self).upper()
    
    def to_json(self) -> dict[str, Any]:
        return {
            self.NAME_KEY: self.name,
            self.VAR_NAME_KEY: self.var_name_object.to_json()
        }


import os, re

from typing import Any

from .interfaces.IVarName import IVarName
from .interfaces.IVarNameStorageTable import IVarNameStorageTable
from .interfaces.IVarNameStorageContainer import IVarNameStorageContainer

class VarName(IVarName):
    def __init__(self, value: str | dict[Any, Any]):
        self.value = value
    
    def handle_category_prefix(self, var_name_props: dict[Any, Any], category_name: str) -> str:
        if 'prefixCategoryName' in var_name_props:
            # Print debugging output if DEBUG_MODE is set
            if os.getenv('DEBUG_MODE') != None:
                print('The prefix property of the varName property was set')

            # Note, we only want to prefix if it's set to true
            if var_name_props['prefixCategoryName'] == True:
                # Print debugging output if DEBUG_MODE is set
                if os.getenv('DEBUG_MODE') != None:
                    print('The prefix property of the varName property was set to true')

                return category_name
            else:
                return ''
        else:
            # Print debugging output if DEBUG_MODE is set
            if os.getenv('DEBUG_MODE') != None:
                print('The prefix property of the varName property was not set defaulting...');

            # Because the prefixCategoryName property isn't set default to prefixing the category name
            return category_name
    
    def handle_replacements(self, var_name_props: dict[Any, Any], var_name_so_far: str) -> str:
        """Perform the replacements specified in the variable name's property to the encapsulating object's name

        Args:
            var_name_props (dict): The variable name's properties (where the replacements would be set, if set)
            var_name_so_far (str): The variable name so far (the name of the encapsulating object with any prefixes if set)

        Returns:
            str: The variable name with replacements filled in, if set. Otherwise the variable name as is
        """

        if 'replace' in var_name_props and len(var_name_props['replace']) > 0:
            # Print debugging output if DEBUG_MODE is set
            if os.getenv('DEBUG_MODE') != None:
                print('The replace property is set on the varName property object')

            # Variable to hold the results of the replacements as we go
            name_with_replacements = var_name_so_far

            # Loop over the desired replacements and "calculate" the output
            for search_str in var_name_props['replace']:
                # Print debugging output if DEBUG_MODE is set
                if os.getenv('DEBUG_MODE') != None:
                    print(f'Attempting to replace {search_str} with {var_name_props["replace"][search_str]} in {name_with_replacements}')

                # Actually do the replacements
                name_with_replacements = name_with_replacements.replace(search_str, var_name_props['replace'][search_str])

                # Print debugging output if DEBUG_MODE is set
                if os.getenv('DEBUG_MODE') != None:
                    print(f'After replacement {name_with_replacements}')

            # append the name with replacements to the actual full environment variable name being created
            return name_with_replacements
        else:
            # Because no replacements were specified simply use the name of the encapsulating object
            return var_name_so_far

    def handle_suffix(self, var_name_props: dict[Any, Any], var_name_so_far: str) -> str:
        """Provide the suffix if set

        Args:
            var_name_props (dict): The variable name properties (where the suffix would be set, if set)
            encapsulating_obj (dict): The encapsulating object (the object the variable name is for)

        Returns:
            str: The suffix to use, if set. Otherwise an empty string
        """

        # Append the suffix if set
        if 'suffix' in var_name_props:
            suffix = var_name_props['suffix']
            
            # Print debugging output if DEBUG_MODE is set
            if os.getenv('DEBUG_MODE') != None:
                print(f'Adding suffix: ${suffix}')
                    
            return var_name_props['suffix']
        else:
            return ''
    
    def handle_dictionary_value(self, var_name_props: dict[Any, Any], encapsulating_obj: IVarNameStorageTable | IVarNameStorageContainer) -> str:
        env_var_name = ''

        # Print debugging output if DEBUG_MODE is set
        if os.getenv('DEBUG_MODE') != None:
            print('The varName property was an object')
                    
        # Prefix the category name if the setting isn't set (default) or if it is set, but is set to true
        # That is, ONLY if the `prefixCategoryName` property is set to false do we not prefix the category name
        # This is an attempt to keep environment variable names as unique as possible (as we end up with quite a few)
        env_var_name += self.handle_category_prefix(var_name_props, str(encapsulating_obj.category))

        # Append the encapsulating object's name
        env_var_name += encapsulating_obj.name

        # Handle the replacements if set
        # 
        # Note, we do assignment here instead of appending because we want to replace the existing (which would be the 
        # encapsulating object's name prefixed with the category. Unless the `prefixCategoryName` was explicitly set to false, 
        # in which case it's just the encapsulating object's name) with the replaced version 
        # 
        # If no replacements are set this will simply return the env_var_name as is
        env_var_name = self.handle_replacements(var_name_props, env_var_name)

        # Handle the suffix if set
        env_var_name += self.handle_suffix(var_name_props, env_var_name)

        # Convert from camelCase to snake_case
        env_var_name = re.sub(r'(?<!^)(?=[A-Z])', '_', env_var_name).lower()

        # Append an appropriate suffix based on the type of encapsulating object
        if isinstance(encapsulating_obj, IVarNameStorageTable):
            env_var_name += '_STORAGE_TABLE_NAME'
        elif isinstance(encapsulating_obj, IVarNameStorageContainer): # type: ignore
            env_var_name += '_CONTAINER_NAME'

        # Print debugging output if DEBUG_MODE is set
        if os.getenv('DEBUG_MODE') != None:
            print(f'End result of varName object processing: ${env_var_name}')
        
        return env_var_name
    
    def get(self, encapsulating_obj: IVarNameStorageTable | IVarNameStorageContainer) -> str:
        # Print debugging output if DEBUG_MODE is set
        if os.getenv('DEBUG_MODE') != None:
            print(f'The varName property was set to: {str(self.value)}')

        if not isinstance(self.value, str):
            return self.handle_dictionary_value(self.value, encapsulating_obj)
        else:
            # the varName property is a string literal so just use it as the environment variable name
            return self.value
    
    def to_json(self) -> str | dict[str, Any]:
        return self.value


from abc import ABCMeta, abstractmethod

from .IWebApp import IWebApp

class IWebAppAssociatedResource:
    """An abstract class that defines properties for resources that can be associated with a or a collection of webapp(s)"""

    __metaclass__ = ABCMeta

    WEB_APPS_KEY = "webApps"
    """The key for the associated web apps array of the resource in the definition JSON"""

    @property
    @abstractmethod
    def web_apps(self) -> list[IWebApp]:
        pass

    


from abc import ABCMeta, abstractmethod

from .ICategorizableResource import ICategorizableResource
from .IWebAppAssociatedResource import IWebAppAssociatedResource
from .IStorageContainer import IStorageContainer
from .IStorageTable import IStorageTable

class IStorageAccount(ICategorizableResource, IWebAppAssociatedResource):
    """An abstract representation of the internal representation of an Azure Storage Account"""

    __metaclass__ = ABCMeta
    
    CONTAINERS_KEY = "containers"
    """The key for the containers of the storage account in the definition JSON"""
    TABLES_KEY = "tables"
    """The key for the tables of the storage account in the definition JSON"""

    @property
    @abstractmethod
    def key(self) -> str:
        """The key of the storage account
        
        Note, this comes from a seperate "storage accounts" file rather than the "resources" file
        """

        pass
    @key.setter
    @abstractmethod
    def key(self, value: str):
        pass
    
    @property
    @abstractmethod
    def connection_string(self) -> str:
        """The connection string for the storage account
        
        Note, this comes from a separate "storage accounts" file rather than the "resources" file
        """

        pass
    @connection_string.setter
    @abstractmethod    
    def connection_string(self, value: str):
        pass

    @property
    @abstractmethod
    def containers(self) -> list[IStorageContainer]:
        """The containers in the storage account"""
        
        pass
    @abstractmethod
    def add_container(self, container: IStorageContainer):
        """Add a container to the storage account

        Args:
            container (IStorageContainer): The storage container to add
        """

        pass
    @abstractmethod
    def has_container(self, container_name: str) -> bool:
        """Check if a container exists in the storage account

        Args:
            container_name (str): The name of the container

        Returns:
            bool: True if the container exists, False otherwise
        """

        pass
    @abstractmethod
    def get_container(self, container_name: str) -> IStorageContainer | None:
        """Get a container from the storage account

        Args:
            container_name (str): The name of the container

        Returns:
            IStorageContainer: The storage container
        """

        pass
    @abstractmethod
    def remove_container(self, container: IStorageContainer):
        """Remove a container from the storage account

        Args:
            container (IStorageContainer): The storage container to remove
        """

        pass

    @property
    @abstractmethod
    def tables(self) -> list[IStorageTable]:
        """The tables in the storage account"""
        
        pass
    @abstractmethod
    def add_table(self, table: IStorageTable):
        """Add a table to the storage account

        Args:
            table (IStorageTable): The storage table to add
        """

        pass
    @abstractmethod
    def has_table(self, table_name: str) -> bool:
        """Check if a table exists in the storage account

        Args:
            table_name (str): The name of the table

        Returns:
            bool: True if the table exists, False otherwise
        """

        pass
    @abstractmethod
    def get_table(self, table_name: str) -> IStorageTable | None:
        """Get a table from the storage account

        Args:
            table_name (str): The name of the table

        Returns:
            IStorageTable: The storage table
        """

        pass
    @abstractmethod
    def remove_table(self, table: IStorageTable):
        """Remove a table from the storage account

        Args:
            table (IStorageTable): The storage table to remove
        """

        pass


from abc import ABCMeta

from .IResource import IResource

class ICategorizableResource(IResource):
    """An abstract class that defines properties for any kind of resource that can have a BA infrastructure category"""

    __metaclass__ = ABCMeta

    @property
    def category(self) -> str | None:
        """The BA infrastructure category of the resource

        Returns:
            str | None: The BA infrastructure category of the resource (or None if not applicable)
        """
        
        return None